import importlib
import inspect
import os
import re
import pkgutil

from packaging.specifiers import SpecifierSet
from packaging.version import Version
from typing import Coroutine, Dict, Iterator, List, Tuple
from types import ModuleType, FunctionType

from schema import SchemaError

from starlette.applications import Starlette
from starlette.routing import Router, Route

import yaml


from . import __version__
from .lib.constants import API_SCHEMA_DICT, ROUTER_SCHEMA, VERBS
from .half_route import HalfRoute
from .lib import acl
from .lib.responses import ORJSONResponse
from .lib.routes import JSONRoute
from .lib.domain import MissingAclError, PathError, UnknownPathParameterType, \
    UndefinedRoute, UndefinedFunction, get_fct_name, route_decorator
from .lib.domain_middleware import DomainMiddleware
from .logging import logger

class HalfDomain(Starlette):
    def __init__(self, domain, module=None, router=None, acl=None, app=None):
        """
        Parameters:
            domain (str): Module name (should be importable)
            router (str): Router name (should be importable from domain module
                defaults to __router__ variable from domain module)
            app (HalfAPI): The app instance
        """
        self.app = app

        self.m_domain = importlib.import_module(domain) if module is None else module
        self.name = getattr(self.m_domain, '__name__', domain)
        self.id = getattr(self.m_domain, '__id__')
        self.version = getattr(self.m_domain, '__version__', '0.0.0')
        # TODO: Check if given domain halfapi_version matches with __version__
        self.halfapi_version = getattr(self.m_domain, '__halfapi_version__', __version__)

        self.deps = getattr(self.m_domain, '__deps__', tuple())

        if not router:
            self.router = getattr(domain, '__router__', '.routers')
        else:
            self.router = router

        self.m_router = None
        try:
            self.m_router = importlib.import_module(self.router, self.m_domain.__package__)
        except AttributeError:
            raise Exception('no router module')

        self.m_acl = HalfDomain.m_acl(self.m_domain, acl)

        self.config = { **app.config }

        logger.info('HalfDomain creation %s %s', domain, self.config)

        for elt in self.deps:
            package, version = elt
            specifier = SpecifierSet(version)
            package_module = importlib.import_module(package)
            if Version(package_module.__version__) not in specifier:
                raise Exception(
                    'Wrong version for package {} version {} (excepting {})'.format(
                        package, package_module.__version__, specifier
                    ))


        super().__init__(
            routes=[ elt for elt in self.gen_domain_routes() ],
            middleware=[
                (DomainMiddleware, {
                    'domain': {
                        'name': self.name,
                        'id': self.id,
                        'version': self.version,
                        'halfapi_version': self.halfapi_version,
                        'config': self.config.get('domain', {}).get(self.name, {}).get('config', {})
                    }
                })
            ]
        )

    @staticmethod
    def m_acl(module, acl=None):
        """ Returns the imported acl module for the domain module
        """
        if not acl:
            acl = getattr(module, '__acl__', '.acl')

        return importlib.import_module(acl, module.__package__)


    @staticmethod
    def acls(module, acl=None):
        """ Returns the ACLS constant for the given domain
        """
        m_acl = HalfDomain.m_acl(module, acl)
        try:
            return getattr(m_acl, 'ACLS')
        except AttributeError:
            raise Exception(f'Missing acl.ACLS constant in module {m_acl.__package__}')

    @staticmethod
    def acls_route(domain, module_path=None, acl=None):
        """ Dictionary of acls

        Format :

        {
            [acl_name]: {
                callable: fct_reference,
                docs: fct_docstring,
                result: fct_result
            }
        }
        """

        d_res = {}

        module = importlib.import_module(domain) \
            if module_path is None \
            else importlib.import_module(module_path)

        m_acl = HalfDomain.m_acl(module, acl)

        for acl_name, doc, order in HalfDomain.acls(
            module,
            acl=acl):
            fct = getattr(m_acl, acl_name)
            d_res[acl_name] = {
                'callable': fct,
                'docs': doc,
                'result': None
            }
        return d_res

    # def schema(self):

    @staticmethod
    def gen_routes(m_router: ModuleType,
        verb: str,
        path: List[str],
        params: List[Dict]) -> Tuple[FunctionType, Dict]:
        """
        Returns a tuple of the function associatied to the verb and path arguments,
        and the dictionary of it's acls

        Parameters:
            - m_router (ModuleType): The module containing the function definition

            - verb (str): The HTTP verb for the route (GET, POST, ...)

            - path (List): The route path, as a list (each item being a level of
                deepness), from the lowest level (domain) to the highest

            - params (Dict): The acl list of the following format :
                [{'acl': Function, 'args': {'required': [], 'optional': []}}]


        Returns:

            (Function, Dict): The destination function and the acl dictionary

        """
        if len(params) == 0:
            raise MissingAclError('[{}] {}'.format(verb, '/'.join(path)))

        if len(path) == 0:
            logger.error('Empty path for [{%s}]', verb)
            raise PathError()

        fct_name = get_fct_name(verb, path[-1])
        if hasattr(m_router, fct_name):
            fct = getattr(m_router, fct_name)
        else:
            raise UndefinedFunction('{}.{}'.format(m_router.__name__, fct_name or ''))


        if not inspect.iscoroutinefunction(fct):
            return route_decorator(fct), params

        # TODO: Remove when using only sync functions
        return acl.args_check(fct), params


    @staticmethod
    def gen_router_routes(
            m_router,
            path: List[str]) -> \
        Iterator[Tuple[str, str, ModuleType, Coroutine, List]]:
        """
        Recursive generator that parses a router (or a subrouter)
        and yields from gen_routes

        Parameters:

            - m_router (ModuleType): The currently treated router module
            - path (List[str]): The current path stack

        Yields:

            HalfRoute
        """
        def read_router(m_router: ModuleType, path: List[str]) -> \
            Iterator[HalfRoute]:
            """
            Reads a module and yields the HalfRoute objects
            """
            try:
                yield from (
                    HalfRoute(
                        path,
                        getattr(m_router, verb),
                        verb
                    )
                    for verb in map(str.lower, VERBS)
                    if getattr(m_router, verb, False)
                )
            except AttributeError:
                """ The router has no function for verb
                """
                pass

        for _loader, subpath, is_pkg in pkgutil.walk_packages(m_router.__path__):
            if not is_pkg:
                """ Do not treat if it is not a package
                """
                continue

            param_match = re.fullmatch('^([A-Z_]+)_([a-z]+)$', subpath)
            if param_match is not None:
                try:
                    path.append('{{{}:{}}}'.format(
                        param_match.groups()[0].lower(),
                        param_match.groups()[1]))
                except AssertionError as exc:
                    raise UnknownPathParameterType(subpath) from exc
            else:
                path.append(subpath)

            # yield ('/'.join(filter(lambda x: len(x) > 0, path)),
            #     verb,
            #     m_router,
            #     *HalfDomain.gen_routes(m_router, verb, path, params[verb])
            # )

            try:
                yield from HalfDomain.gen_router_routes(
                    importlib.import_module( f'.{subpath}', m_router.__name__),
                    path
                )

                path.pop()
            except ImportError as exc:
                logger.error('Failed to import subroute **{%s}**', subpath)
                raise exc

        yield from read_router(m_router, path)


    def gen_domain_routes(self):
        """
        Yields the Route objects for a domain

        Parameters:
            m_domains: ModuleType

        Returns:
            Generator(HalfRoute)
        """
        async def route(request, *args, **kwargs):
            return ORJSONResponse([ self.schema() ])

        yield Route(
            path='/',
            endpoint=route,
            methods=['GET']
        )

        yield from HalfDomain.gen_router_routes(self.m_router, [])

    def schema_dict(self, acls=[{'acl': acl.public}]) -> Dict:
        """ gen_router_routes return values as a dict
        Parameters:

        m_router (ModuleType): The domain routers' module

        Returns:

        Dict: Schema of dict is halfapi.lib.constants.DOMAIN_SCHEMA

        @TODO: Should be a "router_schema_dict" function
        """
        d_res = {}

        for half_route in HalfDomain.gen_router_routes(self.m_router, []):
            path = half_route.path
            verb = list(half_route.methods)[0]
            fct = half_route.endpoint.__name__

            if path not in d_res:
                d_res[path] = {}

            if verb not in d_res[path]:
                d_res[path][verb] = {}

            d_res[path][verb]['callable'] = f'{path}:{fct}'
            try:
                d_res[path][verb]['docs'] = yaml.safe_load(fct.__doc__)
            except AttributeError:
                logger.error(
                    'Cannot read docstring from fct (fct=%s path=%s verb=%s', fct.__name__, path, verb)

            d_res[path][verb]['acls'] = list(map(
                lambda elt: { **elt, 'acl': elt['acl'].__name__ },
                half_route.acls))

        return d_res


    def schema(self) -> Dict:
        schema = { **API_SCHEMA_DICT }
        schema['domain'] = {
            'name': self.name,
            'id': self.id,
            'version': getattr(self.m_domain, '__version__', ''),
            'patch_release': getattr(self.m_domain, '__patch_release__', ''),
            'routers': self.m_router.__name__,
            'acls': tuple(getattr(self.m_acl, 'ACLS', ()))
        }
        schema['paths'] = self.schema_dict()
        return schema
