""" HalfRoute

Child class of starlette.routing.Route
"""
import inspect
from functools import partial, wraps

from typing import Callable, Coroutine, List, Dict
from types import FunctionType

from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.exceptions import HTTPException

from .logging import logger
from .lib.domain import MissingAclError, PathError, UnknownPathParameterType, \
    UndefinedRoute, UndefinedFunction, route_decorator

class HalfRoute(Route):
    """ HalfRoute
    """
    def __init__(self, path: List[str], fct: Callable, method: str, acls=[]):
        logger.info('HalfRoute creation: %s %s %s', path, fct, method)

        fct_args_spec = inspect.getfullargspec(fct)
        fct_args_defaults_dict = {}
        fct_args_defaults = inspect.getfullargspec(fct).defaults or []
        for i in range(len(fct_args_defaults)):
            fct_args_defaults_dict[fct_args_spec.args[-i]] = fct_args_defaults[-i]

        if '__acls' in fct_args_defaults_dict:
            self.acls = fct_args_defaults_dict.get('__acls', {}).copy()

        elif '__acls' in fct_args_spec.kwonlyargs:
            self.acls = fct_args_spec.kwonlydefaults.get('__acls', {}).copy()

        else:
            self.acls = acls.copy()

        if 'ret_type' in fct_args_defaults_dict:
            self.ret_type = fct_args_defaults_dict['ret_type']
        else:
            self.ret_type = 'json'

        print(f'HalfRoute {path} {fct_args_spec} {self.ret_type}')


        if len(self.acls) == 0:
            raise MissingAclError(
                'Route function has no acl attached {}:{}'.format(fct.__module__, fct.__name__))


        super().__init__(
            '/'.join([''] + path),
            HalfRoute.acl_decorator(
                route_decorator(fct),
                self.acls
            ),
            methods=[method])

    @staticmethod
    def acl_decorator(fct, acl_spec) -> Coroutine:
        """
        Decorator for async functions that calls pre-conditions functions
        and appends kwargs to the target function


        Parameters:
            fct (Function):
                The function to decorate

            acl_spec:
                ACL specification

        Returns:
            async function
        """
        @wraps(fct)
        async def caller(req: Request, *args, **kwargs):
            print(f'ACL_DECORATOR {fct} {args} {kwargs}')
            for param in acl_spec:
                if param.get('acl'):
                    passed = param['acl'](req, *args, **kwargs)
                    if isinstance(passed, FunctionType):
                        passed = param['acl']()(req, *args, **kwargs)

                    if not passed:
                        logger.debug(
                            'ACL FAIL for current route (%s)', fct)
                        continue

                    # logger.debug(
                    #     'ACL OK for current route (%s - %s)', fct, param.get('acl'))

                    req.scope['acl_pass'] = param['acl'].__name__

                    if 'args' in param:
                        req.scope['args'] = param['args']
                        logger.debug(
                            'Args for current route (%s)', param.get('args'))

                    if 'out' in param:
                        req.scope['out'] = param['out']

                    if 'out' in param:
                        req.scope['out'] = param['out'].copy()

                    if 'check' in req.query_params:
                        return PlainTextResponse(param['acl'].__name__)

                    # logger.debug('acl_decorator %s', param)
                    # logger.debug('calling %s:%s %s %s', fct.__module__, fct.__name__, args, kwargs)
                    return await fct(
                        req, *args,
                        **{
                            **kwargs,
                        })

            if 'check' in req.query_params:
                return PlainTextResponse('')

            raise HTTPException(401)

        return caller
